// -*- c++ -*- 
//
// BeMailAccount
// Copyright (c) 2001, Okada Jun
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
// 
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// 
// 3. The name of author may not be used to endorse or promote products
// derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 


#include <cstdio>
#include <cstring>
#include <ctime>
#include <Autolock.h>
#include <File.h>
#include <Directory.h>
#include <String.h>
#include <Invoker.h>
#include <Alert.h>
#include <Roster.h>
#include <Messenger.h>
#include <MessageRunner.h>
#include <E-mail.h>
#include <Beep.h>
#include "BMADefs.h"
#include "BMAAccountDialog.h"
#include "BMAProperty.h"


#define ACCOUNT "account"
#define ACTIVE  "active"
#define SMTP    "smtp"
#define POP     "pop"
#define CHECK_DAYS       "check_days"
#define CHECK_INTERVAL   "check_interval"
#define CHECK_BEGIN_TIME "check_begin_time"
#define CHECK_END_TIME   "check_end_time"


BMAProperty::BMAProperty(void)
  : BLooper(BMA_PROPERTY_NAME)
  , mNotifier(NULL)
{
  this->Read();
  this->InitContainer();
  BMessage msg(BMA_AUTO_CHECK);
  mAutoCheckRunner = new BMessageRunner(BMessenger(this),
					&msg,
					this->CheckInterval() * 1000000);
}


BMAProperty::~BMAProperty(void)
{
  delete mAutoCheckRunner;
}


void
BMAProperty::Read(void)
{
  BAutolock lock(this);
  BFile file;

  if (file.SetTo(BMA_CONFIG_FILE, B_READ_ONLY) == B_OK)
    mContainer.Unflatten(&file);
}


void
BMAProperty::Write(void)
{
  BAutolock lock(this);
  BFile file;

  ::create_directory(BMA_CONFIG_DIRECTORY, 0755);

  if (file.SetTo(BMA_CONFIG_FILE, B_WRITE_ONLY | B_CREATE_FILE) == B_OK)
    mContainer.Flatten(&file);
}


void
BMAProperty::InitContainer(void)
{
  type_code type;
  int32 count;

  if (mContainer.GetInfo(CHECK_DAYS, &type, &count) != B_OK)
    mContainer.AddInt32(CHECK_DAYS, B_CHECK_NEVER);

  if (mContainer.GetInfo(CHECK_INTERVAL, &type, &count) != B_OK)
    mContainer.AddInt32(CHECK_INTERVAL, BMA_DEFAULT_CHECK_INTERVAL);

  if (mContainer.GetInfo(CHECK_BEGIN_TIME, &type, &count) != B_OK)
    mContainer.AddInt32(CHECK_BEGIN_TIME, BMA_DEFAULT_CHECK_BEGIN_TIME);

  if (mContainer.GetInfo(CHECK_END_TIME, &type, &count) != B_OK)
    mContainer.AddInt32(CHECK_END_TIME, BMA_DEFAULT_CHECK_END_TIME);
}


int32
BMAProperty::CountAccounts(void)
{
  BAutolock lock(this);
  type_code type;
  int32 count;

  mContainer.GetInfo(ACCOUNT, &type, &count);

  return count;
}


const char *
BMAProperty::AccountAt(int32 index)
{
  BAutolock lock(this);
  const char *account;

  mContainer.FindString(ACCOUNT, index, &account);

  return account;
}


int32
BMAProperty::IndexOfAccount(const char *account)
{
  BAutolock lock(this);
  int32 i, num;

  if (! account)
    return -1;

  num = this->CountAccounts();

  for (i = 0; i < num; i++)
    if (::strcmp(account, this->AccountAt(i)) == 0)
      return i;

  return -1;
}


bool
BMAProperty::HasAccount(const char *account)
{
  return this->IndexOfAccount(account) >= 0;
}


void
BMAProperty::AddAccount(const char *account)
{
  BAutolock lock(this);
  mail_pop_account popDummy;
  BString accountPath;

  if (this->HasAccount(account))
    return;

  mContainer.AddString(ACCOUNT, account);
  mContainer.AddData(POP, B_RAW_TYPE, &popDummy, sizeof(popDummy));
  mContainer.AddString(SMTP, "");

  this->StoreAccount(account);

  this->Write();
}
  

void
BMAProperty::RemoveAccount(const char *account)
{
  BAutolock lock(this);
  int32 index;

  index = this->IndexOfAccount(account);
  if (index < 0)
    return;

  mContainer.RemoveData(ACCOUNT, index);
  mContainer.RemoveData(POP, index);
  mContainer.RemoveData(SMTP, index);

  this->Write();
}


void 
BMAProperty::SetCheckDays(int32 data)
{
  BAutolock lock(this);

  mContainer.ReplaceInt32(CHECK_DAYS, data);

  this->Write();
}


int32
BMAProperty::CheckDays(void)
{
  BAutolock lock(this);
  int32 data;

  mContainer.FindInt32(CHECK_DAYS, &data);

  return data;
}


void 
BMAProperty::SetCheckInterval(int32 data)
{
  BAutolock lock(this);

  mContainer.ReplaceInt32(CHECK_INTERVAL, data);

  mAutoCheckRunner->SetInterval(data * 1000000);

  this->Write();
}


int32
BMAProperty::CheckInterval(void)
{
  BAutolock lock(this);
  int32 data;

  mContainer.FindInt32(CHECK_INTERVAL, &data);

  return data;
}


void 
BMAProperty::SetCheckBeginTime(int32 data)
{
  BAutolock lock(this);

  mContainer.ReplaceInt32(CHECK_BEGIN_TIME, data);

  this->Write();
}


int32
BMAProperty::CheckBeginTime(void)
{
  BAutolock lock(this);
  int32 data;

  mContainer.FindInt32(CHECK_BEGIN_TIME, &data);

  return data;
}


void 
BMAProperty::SetCheckEndTime(int32 data)
{
  BAutolock lock(this);

  mContainer.ReplaceInt32(CHECK_END_TIME, data);

  this->Write();
}


int32
BMAProperty::CheckEndTime(void)
{
  BAutolock lock(this);
  int32 data;

  mContainer.FindInt32(CHECK_END_TIME, &data);

  return data;
}


void
BMAProperty::SetNotifier(BHandler *notifier)
{
  mNotifier = notifier;
}


void
BMAProperty::RestoreAccount(const char *account)
{
  int32 index;
  mail_pop_account pop;
  char smtp[ B_MAX_HOST_NAME_LENGTH ];
  const void *data;
  ssize_t size;

  index = this->IndexOfAccount(account);
  if (index < -1)
    return;

  mContainer.FindData(POP, B_RAW_TYPE, index, &data, &size);
  ::memcpy(&pop, data, size);
  mContainer.FindData(SMTP, B_STRING_TYPE, index, &data, &size);
  ::memcpy(smtp, data, size);

  ::set_pop_account(&pop);
  ::set_smtp_host(smtp);
}


void
BMAProperty::StoreAccount(const char *account)
{
  int32 index;
  mail_pop_account pop;
  char smtp[ B_MAX_HOST_NAME_LENGTH ];

  index = this->IndexOfAccount(account);
  if (index < -1)
    return;

  ::get_pop_account(&pop);
  ::get_smtp_host(smtp);

  mContainer.ReplaceData(POP, B_RAW_TYPE, index, &pop, sizeof(pop));
  mContainer.ReplaceString(SMTP, index, smtp);

  this->Write();
}


void
BMAProperty::SetActiveAccount(const char *newAccount)
{
  BAutolock lock(this);
  const char *oldAccount;
  bool isEmailPrefsAlive;

  oldAccount = this->ActiveAccount();

  if (oldAccount &&
      newAccount &&
      ::strcmp(oldAccount, newAccount) == 0)
    return;
 
  isEmailPrefsAlive = this->IsEMailPrefsAlive();
  if (isEmailPrefsAlive)
    this->QuitEMailPrefs();

  if (oldAccount)
    this->StoreAccount(oldAccount);

  if (newAccount)
    if (oldAccount)
      mContainer.ReplaceString(ACTIVE, newAccount);
    else
      mContainer.AddString(ACTIVE, newAccount);
  else
    mContainer.RemoveName(ACTIVE);

  if (newAccount)
    this->RestoreAccount(newAccount);

  this->Write();

  if (isEmailPrefsAlive)
    this->LaunchEMailPrefs();
}


const char *
BMAProperty::ActiveAccount(void)
{
  BAutolock lock(this);
  const char *account;

  mContainer.FindString(ACTIVE, &account);

  return account;
}


bool
BMAProperty::IsEMailPrefsAlive(void)
{
  return be_roster->IsRunning(EMAIL_PREFS_SIGNATURE);
}


void
BMAProperty::QuitEMailPrefs(void)
{
  BMessenger emailPrefs(EMAIL_PREFS_SIGNATURE);

  if (! emailPrefs.IsValid())
    return;

  emailPrefs.SendMessage(B_QUIT_REQUESTED);

  while (be_roster->IsRunning(EMAIL_PREFS_SIGNATURE))
    ::snooze(10000);
}


void
BMAProperty::LaunchEMailPrefs(void)
{
  be_roster->Launch(EMAIL_PREFS_SIGNATURE);
}


#define POP_LENGTH B_MAX_USER_NAME_LENGTH + \
                   B_MAX_USER_NAME_LENGTH + \
                   B_MAX_HOST_NAME_LENGTH

void
BMAProperty::CheckNow(void)
{
  int32 i, j, countSum, count, activeIndex;
  mail_pop_account popBackup, pop;
  mail_notification notify;
  const void *data, *data2;
  ssize_t size;
  BString err;

  ::get_pop_account(&popBackup);
  ::get_pop_account(&pop);

  if (pop.days != B_CHECK_NEVER)
    {
      BAlert *alert;
      alert = new BAlert("Warning",
			 "BeMailAccountM Error:\n"
			 "Set E-mail preferences \"Check mail\" "
			 "item to \"Never\" before Check E-mails.",
			 "Ok", NULL, NULL, B_WIDTH_AS_USUAL, B_WARNING_ALERT);
      ::system_beep("Beep");
      alert->Go();
      return;
    }

  BMessenger(mNotifier).SendMessage(BMA_BEGIN_CHECK);

  countSum = 0;
  activeIndex = this->IndexOfAccount(this->ActiveAccount());
  if (activeIndex >= 0)
    {
      mContainer.FindData(POP, B_RAW_TYPE, activeIndex, &data, &size);
      if (::memcmp(&pop, data, sizeof(mail_pop_account)) != 0)
	this->StoreAccount(this->ActiveAccount());
    }

  for (i = 0; i < this->CountAccounts(); i++)
    {
      const mail_pop_account *pop1, *pop2;
      mContainer.FindData(POP, B_RAW_TYPE, i, &data, &size);
      pop1 = reinterpret_cast<const mail_pop_account*>(data);

      for (j = 0; j < i; j++)
	{
	  mContainer.FindData(POP, B_RAW_TYPE, j, &data2, &size);
	  pop2 = reinterpret_cast<const mail_pop_account*>(data2);
	  if (::strcmp(pop1->pop_name, pop2->pop_name) == 0 &&
	      ::strcmp(pop1->pop_password, pop2->pop_password) == 0 && 
	      ::strcmp(pop1->pop_host, pop2->pop_host) == 0)
	    break;
	}

      if (j != i)
	continue;

      ::memcpy(&pop, data, POP_LENGTH);
      ::set_pop_account(&pop);

      switch (::check_for_mail(&count))
	{
	default:
	case B_OK:
	  countSum += count;
	  break;
	  
	case B_MAIL_NO_DAEMON:
	  err << "mail_daemon isn't running.\n";
	  break;
	  
	case B_MAIL_UNKNOWN_USER:
	  err << "POP server\""
	      << pop.pop_host
	      << "\" doesn't recognize user name \""
	      << pop.pop_name
	      << "\".\n";
	  break;
	  
	case B_MAIL_WRONG_PASSWORD:
	  err << "POP server\""
	      << pop.pop_host
	      << "\" doesn't recognize the password.\n";
	  break;
	  
	case B_MAIL_UNKNOWN_HOST:
	  err << "POP server\""
	      << pop.pop_host
	      << "\" can't be found.\n";
	  break;
	  
	case B_MAIL_ACCESS_ERROR:
	  err << "The connection to POP server\""
	      << pop.pop_host
	      << "\" failed.\n";
	  break;
	}
    }

  ::set_pop_account(&popBackup);

  ::get_mail_notification(&notify);

  if (countSum && notify.beep)
    ::system_beep("New E-mail");

  if (err.Length())
    ::system_beep("Beep");

  BMessenger(mNotifier).SendMessage(BMA_END_CHECK);
}


void
BMAProperty::AutoCheck(void)
{
  time_t now;
  tm *t;

  now = ::time(NULL);
  t = ::localtime(&now);

  switch (this->CheckDays())
    {
    case B_CHECK_NEVER:
      goto autoCheckSkip;

    case B_CHECK_WEEKDAYS:
      if (t->tm_wday == 0 || t->tm_wday == 6)
	goto autoCheckSkip;
      // THROUGH

    case B_CHECK_DAILY:
      {
	int32 beginTime, endTime, sec;
	beginTime = this->CheckBeginTime();
	endTime = this->CheckEndTime();
	sec = t->tm_sec + 60 * t->tm_min + 3600 * t->tm_hour;
      if (beginTime < endTime)
	{
	  if (sec < beginTime || endTime < sec)
	    goto autoCheckSkip;
	}
      else
	{
	  if (beginTime < sec && sec < endTime)
	    goto autoCheckSkip;
	}
      break;
      }

    case B_CHECK_CONTINUOUSLY:
    default:
      break;
    }

  this->CheckNow();

 autoCheckSkip:
  ;
}


void
BMAProperty::MessageReceived(BMessage *msg)
{
  switch (msg->what)
    {
    default:
      BLooper::MessageReceived(msg);
      break;

    case BMA_SET_ACTIVE_ACCOUNT:
      {
	const char *account;

	msg->FindString("data", &account);
	this->SetActiveAccount(account);
	break;
      }
      
    case BMA_NEW_ACCOUNT_REQUESTED:
      {
	BMAAccountDialog *dialog;

	dialog = new BMAAccountDialog(this, new BMessage(BMA_NEW_ACCOUNT));
	dialog->Go();
	break;
      }

    case BMA_NEW_ACCOUNT:
      {
	const char *account;

	msg->FindString("data", &account);

	if (! *account)
	  break;

	if (this->HasAccount(account))
	  break;

	this->AddAccount(account);
	this->SetActiveAccount(account);

	break;
      }	  

    case BMA_REMOVE_ACCOUNT_REQUESTED:
      {
	BMessage *execMsg;
	BAlert *alert;
	BInvoker *invoker;
	BString buf;
	const char *account;

	execMsg = new BMessage(*msg);
	execMsg->what = BMA_REMOVE_ACCOUNT;

	invoker = new BInvoker(execMsg, this);

	msg->FindString("data", &account);
	buf << "Really Remove \"" << account << "\" ?";
	alert = new BAlert("Remove Account",
			   buf.String(), 
			   "Ok",
			   "Cancel",
			   NULL,
			   B_WIDTH_AS_USUAL);

	alert->Go(invoker);

	break;
      }

    case BMA_REMOVE_ACCOUNT:
      {
	const char *account;
	int32 which;

	msg->FindInt32("which", &which);
	if (which != 0)
	  break;

	msg->FindString("data", &account);
	this->SetActiveAccount(NULL);
	this->RemoveAccount(account);
	
	break;
      }

    case BMA_LAUNCH_EMAIL_PREFS:
      {
	this->LaunchEMailPrefs();
	break;
      }

    case BMA_SET_CHECK_DAYS:
      {
	int32 data;
	msg->FindInt32("data", &data);
	this->SetCheckDays(data);
	break;
      }

    case BMA_SET_CHECK_INTERVAL:
      {
	int32 data;
	msg->FindInt32("data", &data);
	this->SetCheckInterval(data);
	break;
      }

    case BMA_SET_CHECK_BEGIN_TIME:
      {
	int32 data;
	msg->FindInt32("data", &data);
	this->SetCheckBeginTime(data);
	break;
      }

    case BMA_SET_CHECK_END_TIME:
      {
	int32 data;
	msg->FindInt32("data", &data);
	this->SetCheckEndTime(data);
	break;
      }

    case BMA_CHECK_NOW:
      {
	this->CheckNow();
	break;
      }

    case BMA_AUTO_CHECK:
      {
	this->AutoCheck();
	break;
      }
    }
}



